@@ -877,22 +877,6 @@ z_py_ssl_privkey_free(gpointer value)
   EVP_PKEY_free(*pkey);
 }
 
-ZPolicyObj *
-z_py_ssl_cert_list_get(ZProxy * /* self */, gchar * /* name */, gpointer value)
-{
-  STACK_OF(X509) **certlist = (STACK_OF(X509) **) value;
-
-  return z_py_zorp_cert_list_new(*certlist);
-}
-
-void
-z_py_ssl_cert_list_free(gpointer value)
-{
-  STACK_OF(X509) **certlist = (STACK_OF(X509) **) value;
-
-  sk_X509_pop_free(*certlist, X509_free);
-}
-
 ZPolicyObj *
 z_py_ssl_cert_name_list_get(ZProxy * /* self */, gchar * /* name */, gpointer value)
 {
@@ -908,19 +892,3 @@ z_py_ssl_cert_name_list_free(gpointer value)
 
   sk_X509_NAME_pop_free(*certnamelist, X509_NAME_free);
 }
-
-ZPolicyObj *
-z_py_ssl_crl_list_get(ZProxy * /* self */, gchar * /* name */, gpointer value)
-{
-  STACK_OF(X509_CRL) **crllist = (STACK_OF(X509_CRL) **) value;
-
-  return z_py_zorp_crl_list_new(*crllist);
-}
-
-void
-z_py_ssl_crl_list_free(gpointer value)
-{
-  STACK_OF(X509_CRL) **crllist = (STACK_OF(X509_CRL) **) value;
-
-  sk_X509_CRL_pop_free(*crllist, X509_CRL_free);
-}
